@@ -23,6 +23,7 @@ __date__ = "$Sep 08, 2016 15:46:46 EDT$"
 
 
 import itertools
+import numbers
 import operator
 import math
 import warnings
@@ -44,144 +45,105 @@ def reformat_slice(a_slice, a_length=None):
                                    possible.
 
         Examples:
-            >>> reformat_slice(slice(None))
-            slice(0, None, 1)
 
-            >>> reformat_slice(slice(None), 10)
-            slice(0, 10, 1)
-
-            >>> reformat_slice(slice(2, None))
-            slice(2, None, 1)
-
-            >>> reformat_slice(slice(2, None), 10)
-            slice(2, 10, 1)
-
-            >>> reformat_slice(slice(2, None, None))
-            slice(2, None, 1)
-
-            >>> reformat_slice(slice(2, None, None), 10)
-            slice(2, 10, 1)
+            >>> reformat_slice(slice(2, -1, None))
+            slice(2, -1, 1)
 
             >>> reformat_slice(slice(2, -1, None), 10)
             slice(2, 9, 1)
-
-            >>> range(10)[reformat_slice(slice(None))] == range(10)[:]
-            True
-
-            >>> range(10)[reformat_slice(slice(2, None))] == range(10)[2:]
-            True
-
-            >>> range(10)[reformat_slice(slice(2, 6))] == range(10)[2:6]
-            True
-
-            >>> range(10)[reformat_slice(slice(2, 6, 3))] == range(10)[2:6:3]
-            True
-
-            >>> range(10)[reformat_slice(slice(2, None, 3))] == range(10)[2::3]
-            True
-
-            >>> range(10)[reformat_slice(slice(None), 10)] == range(10)[:]
-            True
-
-            >>> range(10)[reformat_slice(slice(2, None), 10)] == range(10)[2:]
-            True
-
-            >>> range(10)[reformat_slice(slice(2, 6), 10)] == range(10)[2:6]
-            True
-
-            >>> range(10)[reformat_slice(slice(2, 6, 3), 10)] == range(10)[2:6:3]
-            True
-
-            >>> range(10)[reformat_slice(slice(2, None, 3), 10)] == range(10)[2::3]
-            True
-
-            >>> range(10)[reformat_slice(slice(2, -6, 3), 10)] == range(10)[2:-6:3]
-            True
-
-            >>> range(10)[reformat_slice(slice(2, -1), 10)] == range(10)[2:-1]
-            True
-
-            >>> range(10)[reformat_slice(slice(2, 20), 10)] == range(10)[2:20]
-            True
-
-            >>> range(10)[reformat_slice(slice(2, -20), 10)] == range(10)[2:-20]
-            True
-
-            >>> range(10)[reformat_slice(slice(20, -1), 10)] == range(10)[20:-1]
-            True
-
-            >>> range(10)[reformat_slice(slice(-20, -1), 10)] == range(10)[-20:-1]
-            True
-
-            >>> range(10)[reformat_slice(slice(-5, -1), 10)] == range(10)[-5:-1]
-            True
     """
 
-    assert (a_slice is not None), "err"
+    new_slice = a_slice
+    if (new_slice is Ellipsis) or (new_slice == tuple()):
+        new_slice = slice(None)
+    elif isinstance(a_slice, numbers.Integral):
+        if a_slice < 0:
+            new_slice = slice(a_slice, a_slice-1, -1)
+        else:
+            new_slice = slice(a_slice, a_slice+1, 1)
+    elif not isinstance(a_slice, slice):
+        raise ValueError(
+            "Expected a `slice` type. Instead got `%s`." % str(a_slice)
+        )
+
+    if new_slice.step == 0:
+        raise ValueError("Slice cannot have a step size of `0`.")
+
+    start = new_slice.start
+    stop = new_slice.stop
+    step = new_slice.step
 
     # Fill unknown values.
+    if step is None:
+        step = 1
+    if start is None:
+        if step > 0:
+            start = 0
+        elif step < 0:
+            start = -1
+    if (stop is None) and (step > 0):
+        stop = a_length
 
-    new_slice_step = a_slice.step
-    if new_slice_step is None:
-        new_slice_step = 1
-
-    new_slice_start = a_slice.start
-    if (new_slice_start is None) and (new_slice_step > 0):
-        new_slice_start = 0
-
-    new_slice_stop = a_slice.stop
-
+    stop_i = stop is not None
 
     # Make adjustments for length
-
     if a_length is not None:
-        if (new_slice_step < -a_length):
-            new_slice_step = -a_length
-        elif (new_slice_step > a_length):
-            new_slice_step = a_length
+        # Normalize out-of-bound step sizes.
+        if step < -a_length:
+            step = -a_length
+        elif step > a_length:
+            step = a_length
 
-        if (new_slice_start is None) and (new_slice_step > 0):
-            pass
-        elif (new_slice_start is None) and (new_slice_step < 0):
-            new_slice_start = a_length
-        elif (new_slice_start <= -a_length) and (new_slice_step > 0):
-            new_slice_start = 0
-        elif (new_slice_start < -a_length) and (new_slice_step < 0):
-            new_slice_start = new_slice_stop = 0
-        elif (new_slice_start > a_length) and (new_slice_step > 0):
-            new_slice_start = a_length
-        elif (new_slice_start > a_length) and (new_slice_step < 0):
-            new_slice_start = a_length
-        elif (new_slice_start < 0) and (new_slice_step > 0):
-            new_slice_start += a_length
-        elif (new_slice_start < 0) and (new_slice_step < 0):
-            new_slice_start += a_length
+        # Normalize bounded negative values.
+        if -a_length <= start < 0:
+            start += a_length
+        if stop_i and (-a_length <= stop < 0):
+            stop += a_length
 
-        if (new_slice_stop is None) and (new_slice_step > 0):
-            new_slice_stop = a_length
-        elif (new_slice_stop is None) and (new_slice_step < 0):
-            pass
-        elif (new_slice_stop <= -a_length) and (new_slice_step > 0):
-            new_slice_stop = 0
-        elif (new_slice_stop < -a_length) and (new_slice_step < 0):
-            new_slice_stop = None
-        elif (new_slice_stop < 0) and (new_slice_step > 0):
-            new_slice_stop += a_length
-        elif (new_slice_stop < 0) and (new_slice_step < 0):
-            new_slice_stop += a_length
-        elif (new_slice_stop > a_length) and (new_slice_step > 0):
-            new_slice_stop = a_length
-        elif (new_slice_stop >= a_length) and (new_slice_step < 0):
-            new_slice_start = new_slice_stop = 0
-        elif (new_slice_stop < 0) and (new_slice_step > 0):
-            new_slice_stop += a_length
-        elif (new_slice_stop < 0) and (new_slice_step < 0):
-            new_slice_stop += a_length
+        # Handle out-of-bound limits.
+        if step > 0:
+            if (start > a_length) or (stop < -a_length):
+                start = stop = 0
+                step = 1
+            else:
+                if start < -a_length:
+                    start = 0
+                if stop > a_length:
+                    stop = a_length
+        elif step < 0:
+            if (start < -a_length) or (stop_i and stop >= (a_length - 1)):
+                start = stop = 0
+                step = 1
+            else:
+                if start >= a_length:
+                    start = a_length - 1
+                if stop_i and stop < -a_length:
+                    stop = None
+                    stop_i = False
 
+    # Catch some known empty slices.
+    if stop_i and (start == stop):
+        start = stop = 0
+        step = 1
+    elif (step > 0) and (stop == 0):
+        start = stop = 0
+        step = 1
+    elif (step < 0) and (stop == -1):
+        start = stop = 0
+        step = 1
+    elif stop_i and (start >= 0) and (stop >= 0):
+        if (step > 0) and (start > stop):
+            start = stop = 0
+            step = 1
+        elif (step < 0) and (start < stop):
+            start = stop = 0
+            step = 1
 
-    # Build new slice and return.
-
-    new_slice = slice(new_slice_start, new_slice_stop, new_slice_step)
+    new_slice = slice(start, stop, step)
+    if isinstance(a_slice, numbers.Integral):
+        if new_slice.start == new_slice.stop == 0:
+            raise IndexError("Index out of range.")
+        new_slice = new_slice.start
 
     return(new_slice)
 
@@ -201,20 +163,6 @@ def reformat_slices(slices, lengths=None):
 
         Examples:
 
-            >>> reformat_slices(slice(None))
-            (slice(0, None, 1),)
-
-            >>> reformat_slices((slice(None),))
-            (slice(0, None, 1),)
-
-            >>> reformat_slices((
-            ...     slice(None),
-            ...     slice(3, None),
-            ...     slice(None, 5),
-            ...     slice(None, None, 2)
-            ... ))
-            (slice(0, None, 1), slice(3, None, 1), slice(0, 5, 1), slice(0, None, 2))
-
             >>> reformat_slices(
             ...     (
             ...         slice(None),
@@ -227,27 +175,89 @@ def reformat_slices(slices, lengths=None):
             (slice(0, 10, 1), slice(3, 13, 1), slice(0, 5, 1), slice(0, 20, 2))
     """
 
+    new_slices = slices
+    if new_slices == tuple():
+        new_slices = Ellipsis
+
     try:
-        len(slices)
+        len(new_slices)
     except TypeError:
-        slices = (slices,)
+        new_slices = (new_slices,)
 
     new_lengths = lengths
-    if new_lengths is None:
-        new_lengths = [None] * len(slices)
-
     try:
-        len(new_lengths)
+        if new_lengths is not None:
+            len(new_lengths)
     except TypeError:
         new_lengths = (new_lengths,)
 
-    assert (len(slices) == len(new_lengths))
+    el_idx = None
+    try:
+        el_idx = new_slices.index(Ellipsis)
+    except ValueError:
+        pass
 
-    new_slices = list(slices)
-    for i, each_length in enumerate(new_lengths):
-        new_slices[i] = reformat_slice(new_slices[i], each_length)
+    if new_lengths is not None and el_idx is None:
+        if len(new_slices) != len(new_lengths):
+            raise ValueError("Shape must be the same as the number of slices.")
+    elif new_lengths is not None:
+        if (len(new_slices) - 1) > len(new_lengths):
+            raise ValueError(
+                "Shape must be as large or larger than the number of slices"
+                " without the Ellipsis."
+            )
 
-    new_slices = tuple(new_slices)
+    if el_idx is not None:
+        # Break into three cases.
+        #
+        # 1. Before the Ellipsis
+        # 2. The Ellipsis
+        # 3. After the Ellipsis
+        #
+        # Cases 1 and 3 are trivially solved as before.
+        # Case 2 is either a no-op or a bunch of `slice(None)`s.
+        #
+        # The result is a combination of all of these.
+
+        slices_before = new_slices[:el_idx]
+        slices_after = new_slices[el_idx+1:]
+
+        if Ellipsis in slices_before or Ellipsis in slices_after:
+            raise ValueError("Only one Ellipsis is permitted. Found multiple.")
+
+        new_lengths_before = None
+        new_lengths_after = None
+        slice_el = (Ellipsis,)
+        if new_lengths is not None:
+            pos_before = len(slices_before)
+            pos_after = len(new_lengths) - len(slices_after)
+
+            new_lengths_before = new_lengths[:pos_before]
+            new_lengths_after = new_lengths[pos_after:]
+
+            new_lengths_el = new_lengths[pos_before:pos_after]
+            slice_el = len(new_lengths_el) * (slice(None),)
+            if slice_el:
+                slice_el = reformat_slices(
+                    slice_el,
+                    new_lengths_el
+                )
+
+        if slices_before:
+            slices_before = reformat_slices(slices_before, new_lengths_before)
+        if slices_after:
+            slices_after = reformat_slices(slices_after, new_lengths_after)
+
+        new_slices = slices_before + slice_el + slices_after
+    else:
+        if new_lengths is None:
+            new_lengths = [None] * len(new_slices)
+
+        new_slices = list(new_slices)
+        for i, each_length in enumerate(new_lengths):
+            new_slices[i] = reformat_slice(new_slices[i], each_length)
+
+        new_slices = tuple(new_slices)
 
     return(new_slices)
 
@@ -279,51 +289,29 @@ def len_slice(a_slice, a_length=None):
                                    possible.
 
         Examples:
-            >>> len_slice(slice(None)) #doctest: +IGNORE_EXCEPTION_DETAIL
-            Traceback (most recent call last):
-            UnknownSliceLengthException: Cannot determine slice length without a defined end point. The reformatted slice was slice(0, None, 1).
-
-            >>> len_slice(slice(None), 10)
-            10
-
-            >>> len_slice(slice(None), 10) == len(range(10)[:])
-            True
 
             >>> len_slice(slice(2, None), 10)
             8
 
-            >>> len_slice(slice(2, None), 10) == len(range(10)[2:])
-            True
-
-            >>> len_slice(slice(2, None, None), 10)
-            8
-
-            >>> len_slice(slice(2, None, None), 10) == len(range(10)[2:])
-            True
-
             >>> len_slice(slice(2, 6))
             4
-
-            >>> len_slice(slice(2, 6), 1000)
-            4
-
-            >>> len_slice(slice(2, 6), 10) == len(range(10)[2:6])
-            True
-
-            >>> len_slice(slice(2, 6, 3))
-            2
-
-            >>> len_slice(slice(2, 6, 3), 10) == len(range(10)[2:6:3])
-            True
     """
 
+    if isinstance(a_slice, numbers.Integral):
+        raise TypeError(
+            "An integral index does not provide an object with a length."
+        )
+
     new_slice = reformat_slice(a_slice, a_length)
 
     if new_slice.stop is None:
-        raise UnknownSliceLengthException(
-            "Cannot determine slice length without a defined end point. " +
-            "The reformatted slice was " + repr(new_slice) + "."
-        )
+        if new_slice.step > 0:
+            raise UnknownSliceLengthException(
+                "Cannot determine slice length without a defined end point. " +
+                "The reformatted slice was " + repr(new_slice) + "."
+            )
+        else:
+            new_slice = slice(new_slice.start, -1, new_slice.step)
 
     new_slice_diff = float(new_slice.stop - new_slice.start)
 
@@ -346,14 +334,6 @@ def len_slices(slices, lengths=None):
                                          values filled if possible.
 
         Examples:
-            >>> len_slices((
-            ...     slice(None),
-            ...     slice(3, None),
-            ...     slice(None, 5),
-            ...     slice(None, None, 2)
-            ... )) #doctest: +IGNORE_EXCEPTION_DETAIL
-            Traceback (most recent call last):
-            UnknownSliceLengthException: Cannot determine slice length without a defined end point. The reformatted slice was slice(0, None, 1).
 
             >>> len_slices(
             ...     (
@@ -372,7 +352,8 @@ def len_slices(slices, lengths=None):
     lens = []
 
     for each_slice in new_slices:
-        lens.append(len_slice(each_slice))
+        if not isinstance(each_slice, numbers.Integral):
+            lens.append(len_slice(each_slice))
 
     lens = tuple(lens)
 
@@ -432,37 +413,36 @@ def split_blocks(space_shape, block_shape, block_halo=None):
     """
 
     try:
-        xrange
+        irange = xrange
     except NameError:
-        xrange = range
+        irange = range
 
     try:
-        from itertools import ifilter
-        from itertools import imap
-        from itertools import izip
+        from itertools import ifilter, imap
     except ImportError:
-        ifilter = filter
-        imap = map
-        izip = zip
+        ifilter, imap = filter, map
 
     if block_halo is not None:
-        assert (len(space_shape) == len(block_shape) == len(block_halo)), \
-            "The dimensions of `space_shape`, `block_shape`, and " + \
-            "`block_halo` should be the same."
+        if not (len(space_shape) == len(block_shape) == len(block_halo)):
+            raise ValueError(
+                "The dimensions of `space_shape`, `block_shape`, and"
+                " `block_halo` should be the same."
+            )
     else:
-        assert (len(space_shape) == len(block_shape)), \
-            "The dimensions of `space_shape` and `block_shape` " + \
-            "should be the same."
+        if not (len(space_shape) == len(block_shape)):
+            raise ValueError(
+               "The dimensions of `space_shape` and `block_shape` should be"
+               " the same."
+            )
 
         block_halo = tuple()
-        for i in xrange(len(space_shape)):
+        for i in irange(len(space_shape)):
             block_halo += (0,)
 
     vec_add = lambda a, b: imap(operator.add, a, b)
     vec_sub = lambda a, b: imap(operator.sub, a, b)
 
     vec_mul = lambda a, b: imap(operator.mul, a, b)
-    vec_div = lambda a, b: imap(operator.div, a, b)
     vec_mod = lambda a, b: imap(operator.mod, a, b)
 
     vec_nonzero = lambda a: \
@@ -494,7 +474,7 @@ def split_blocks(space_shape, block_shape, block_halo=None):
     haloed_ranges_per_dim = []
     trimmed_halos_per_dim = []
 
-    for each_dim in xrange(len(space_shape)):
+    for each_dim in irange(len(space_shape)):
         # Construct each block using the block size given. Allow to spill over.
         if block_shape[each_dim] == -1:
             block_shape = (block_shape[:each_dim] +
@@ -503,9 +483,9 @@ def split_blocks(space_shape, block_shape, block_halo=None):
 
         # Generate block ranges.
         a_range = []
-        for i in xrange(2):
+        for i in irange(2):
             offset = i * block_shape[each_dim]
-            this_range = xrange(
+            this_range = irange(
                 offset,
                 offset + space_shape[each_dim],
                 block_shape[each_dim]
@@ -514,7 +494,7 @@ def split_blocks(space_shape, block_shape, block_halo=None):
 
         # Add the halo to each block on both sides.
         a_range_haloed = []
-        for i in xrange(2):
+        for i in irange(2):
             sign = 2 * i - 1
 
             haloed = vec_mul(
@@ -529,7 +509,7 @@ def split_blocks(space_shape, block_shape, block_halo=None):
         # Compute how to trim the halo off of each block.
         # Clip each block to the boundaries.
         a_trimmed_halo = []
-        for i in xrange(2):
+        for i in irange(2):
             trimmed = vec_sub(a_range[i], a_range_haloed[0])
             a_trimmed_halo.append(list(trimmed))
             a_range[i] = list(vec_clip(a_range[i], 0, space_shape[each_dim]))
